from builtins import str
import json
import os
import ntpath
import re
from os import listdir
from importlib import reload
from os.path import isfile, join
import boto3

from cloud_resources import cdn_catalyst
from cloud_resources import credentials as cloud_creds
from cloud_resources import storage_catalyst
from db_samvaad.db_handler import DatabaseHandle
import db_samvaad as db_con
from sentrylogger import sentry_constants

from algowrap import constants as consts
from algowrap import credentials as creds
from algowrap import error_handler as err_hndl
from algowrap import helper as hlpr
from algowrap import sys_utils
from algowrap.sentry_client import sentry_client


def _download_input_files(directory_name, input_files):
    """
    Download input files from S3 to local dir.

    This function extracts the input files from the parameter.json
    and downloads those files from S3 and store in the local dir.

    Parameters
    ----------
    directory_name: String
        Name of local directory to store downloaded input files.

    input_urls: String
        Dict containing remote input_url or List of input_urls.

    Returns
    -------
    input_local_files: Path to saved files in local dir.
    """
    input_local_files = dict()
    for file_key, file_name in list(input_files.items()):
        if (isinstance(file_name, list)):
            for fil_nam in file_name:
                local_result_file = directory_name + consts.SEPARAT + \
                                    os.path.basename(fil_nam.get(consts.PATH_IDFR))
                if file_key not in input_local_files:
                    input_local_files[file_key] = []
                input_local_files[file_key].append(local_result_file)
                hlpr.download_save(fil_nam.get(consts.PATH_IDFR),
                                   fil_nam.get(consts.BUCKET_IDFR),
                                   local_result_file)
                                                                   
        elif not(isinstance(file_name, list)) and hlpr.wildcard_present(file_name):
            s3_key = file_name.get(consts.PATH_IDFR)
            s3_bucket = file_name.get(consts.BUCKET_IDFR)
            base_dir = ntpath.dirname(s3_key)
            base_file = ntpath.basename(s3_key)
            s3_objects = storage_catalyst.get_directory_objects_raw(s3_bucket, base_dir)
            for s3_object in s3_objects:
                pattern = re.compile("^[^\s]" + base_file)
                if pattern.match(s3_object.key):
                    object_key_name = ntpath.basename(s3_object.key)
                    local_result_file = directory_name + consts.SEPARAT + object_key_name
                    s3_key = s3_object.key
                    
                    if file_key not in input_local_files:
                        input_local_files[file_key] = []

                    input_local_files[file_key].append(local_result_file)
                    hlpr.download_save(s3_key, s3_bucket, local_result_file)
        else:
            local_result_file = directory_name + consts.SEPARAT + \
                                os.path.basename(file_name.get(consts.PATH_IDFR))
            if file_key not in input_local_files:
                input_local_files[file_key] = None

            input_local_files[file_key] = local_result_file
            hlpr.download_save(file_name.get(consts.PATH_IDFR), file_name.get(consts.BUCKET_IDFR),
                               local_result_file)
    return input_local_files


def algorithm_wrapper(algorithm_func, parameter_data, input_local_files):
    """
    Runs the given function at setu layer with required data and credentials
    from the ENV.

    Parameters
    ----------
    algorithm_func: function
        The function to run at Setu layer.
    parameter_data: Object
        Any parameter data to be passed to the function.
    input_local_files: Dict
        Contains information regarding the input files to the algorithm present
        on local dir.

    Returns
    -------
    algo_return: Object
        The output of algorithm_func.
    """
    db_con.db.update_heartbeat(creds.ITER_ID, 1)
    algo_return = algorithm_func(parameter_data, input_local_files)
    db_con.db.update_heartbeat(creds.ITER_ID, 100)
    return algo_return


def get_matching_s3_keys(bucket, prefix='', suffix=''):
    """
    Generate the keys in an S3 bucket.

    Parameters
    ----------
    bucket: String
        Name of the S3 bucket
    prefix: String
        Only fetch keys that start with this prefix (optional).
    suffix: String
        Only fetch keys that end with this suffix (optional).

    Yields
    -------
    key: String
        Name of the file to download given a matching prefix
    """
    s3 = boto3.client('s3')
    kwargs = {'Bucket': bucket}

    # If the prefix is a single string (not a tuple of strings), we can
    # do the filtering directly in the S3 API.
    if isinstance(prefix, str):
        kwargs['Prefix'] = prefix

    while True:

        # The S3 API response is a large blob of metadata.
        # 'Contents' contains information about the listed objects.
        resp = s3.list_objects_v2(**kwargs)
        for obj in resp['Contents']:
            key = obj['Key']
            if key.startswith(prefix) and key.endswith(suffix):
                yield key

        # The S3 API is paginated, returning up to 1000 keys at a time.
        # Pass the continuation token into the next response, until we
        # reach the final page (when this field is missing).
        try:
            kwargs['ContinuationToken'] = resp['NextContinuationToken']
        except KeyError:
            break


def _select_and_run_analysis(link_run_analysis_func, parameter_data,
                             input_local_files):
    try:
        return link_run_analysis_func(parameter_data, input_local_files)
    except Exception as e:
        raise err_hndl.AlgorithmError(e)


def run_algorithm(link_run_analysis_func):
    """
    The main function that extracts inforamtion and run the analysis.
    These steps are as follows:
        1. Get the iteration of the message that has to be processed
        2. Get the parameter files from the ietration files
        3. Get the algorithm and the files that are needed to run the algorithm
        4. run the algorithm and upload the files back to s3
        5. Delete the made/downloaded files 

    Parameter
    ---------
    link_run_analysis_func: function
        The method that calls appropriate algorithm to run at Setu layer.
    """
    # Reloading cloud_resources modules to set environment variables on Lambda
    reload(cloud_creds)
    reload(cdn_catalyst)
    base_local_path = creds.MOUNT_FILE_PATH
    iter_id = creds.ITER_ID
    # Adding iteration ID to Sentry Logs
    sentry_client.queue_debug_log(str(consts.ITER_ID + str(iter_id)))
    hlpr.log_tracking_packages()
    # Updating RDS state
    db_con.db = DatabaseHandle()
    db_con.db.update_state(iter_id, consts.TASK_RECV)

    # Getting the parameter.json file and the file name of the parameter_json
    parameter_data = hlpr.get_param_json(iter_id)

    # This is the local directory in which data from a given
    # iteration would be stored, After that the directory
    # is made
    directory_name = base_local_path + consts.SEPARAT + iter_id
    if not os.path.exists(directory_name):
        os.makedirs(directory_name)
    
    input_files = parameter_data[consts.USER_INPUT_IDENTIFIER]
    keyless_input_files = input_files.get('keyless_files')
    
    sentry_client.queue_debug_log(str(input_files))
    input_local_files = dict()
    if keyless_input_files:
        if isinstance(keyless_input_files, dict):
            # downloads all files
            all_files = []
            # Gets keys inside prefix/folder
            for key in get_matching_s3_keys(bucket=keyless_input_files.get('bucket'), prefix=keyless_input_files.get('basepath')):
                if key.count('/') < 2:
                    # download the file
                    hlpr.download_save(key, keyless_input_files.get('bucket'),
                                       directory_name + consts.SEPARAT + os.path.basename(key.split('/')[1]))
                    all_files.append(directory_name + consts.SEPARAT + os.path.basename(key.split('/')[1]))

            all_seen_files = [f for f in listdir(directory_name) if isfile(join(directory_name, f))]

            input_local_files.update({'keyless_files': all_files})
        else:
            input_local_files.update(_download_input_files(directory_name,
                                     input_files))
    else:
        input_local_files.update(_download_input_files(directory_name,
                                 input_files))
    # updating in RDS that the task is going to start
    db_con.db.update_state(iter_id, consts.START_TASK)
    # Running the analysis
    algo_out = _select_and_run_analysis(link_run_analysis_func,
                                        parameter_data,
                                        input_local_files)
    # Getting the desired output data format
    output_file_types = hlpr.get_iteration_output_key(iter_id, parameter_data)
    sentry_client.queue_debug_log(consts.OUTPUT_FTYPES + str(output_file_types))
    # uploading the file to S3
    hlpr.upload_result_files(output_file_types, algo_out, directory_name, iter_id)
    # Deleting the temp directory that was created
    sys_utils.clear_output_dirs(iter_id)
    # Updating the state of iteration of iter_id in iteration table
    db_con.db.update_state(iter_id, consts.COMP_TASK)